import multiprocessing
import queue

from .batch import Batch
from .enums import NodeType


class Queue:
    def __init__(
            self,
            name: str,
            node_type: NodeType = NodeType.THREAD,
            maxsize: int = 0
    ):
        self._name = name

        if node_type == NodeType.THREAD:
            self._queue = queue.Queue(maxsize=maxsize)
        elif node_type == NodeType.PROCESS:
            self._queue = multiprocessing.Queue(maxsize=maxsize)

    @property
    def name(self) -> str:
        return self._name

    def empty(self) -> bool:
        return self._queue.empty()

    def put(self, item: Batch) -> None:
        self._queue.put(item)

    def get(self) -> Batch:
        return self._queue.get()



import importlib.resources
import logging
import os
import signal
import time
from datetime import datetime
from statistics import mean, pstdev
from threading import Thread
from typing import List, Optional

import uvicorn
from fastapi import APIRouter, FastAPI, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from .batch import Batch
from .enums import NodeStatus, NodeType
from .node import Node
from .queue import Queue


class Watchdog(Node):
    def __init__(self, **kwargs):
        super().__init__(node_type=NodeType.PROCESS, timeout=10.0, **kwargs)

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        if time.time() - self.shared_memory.get('time', float('inf')) >= 600.0:
            logging.info(f'{self.name}: os.kill(1, signal.SIGTERM)')
            os.kill(1, signal.SIGTERM)

        return None


class Pipeline(Node):
    def __init__(
            self,
            nodes: List[Node],
            *,
            queue_maxsize: int = 16,
            verbose_metrics: bool = False,
            metrics_host: str = '0.0.0.0',
            metrics_port: int = 8080,
            **kwargs
    ):
        super().__init__(timeout=1.0, **kwargs)
        logging.getLogger().setLevel(logging.INFO)
        self._nodes = nodes
        self._queue_maxsize = queue_maxsize
        self._build_pipeline()
        self.start()
        Watchdog().start()

        if verbose_metrics:
            self._templates = Jinja2Templates(directory=importlib.resources.files('pipez.resources') / 'templates')
            app = FastAPI()
            app.mount(path='/static',
                      app=StaticFiles(directory=importlib.resources.files('pipez.resources') / 'static', html=True),
                      name='static')
            router = APIRouter()
            router.add_api_route('/metrics_html', self._metrics_html, methods=['GET'], response_class=HTMLResponse)
            router.add_api_route('/metrics_json', self._metrics_json, methods=['GET'])
            app.include_router(router)
            Thread(target=uvicorn.run, kwargs=dict(app=app, host=metrics_host, port=metrics_port)).start()

    def _metrics_html(self, request: Request):
        return self._templates.TemplateResponse('home.html', dict(request=request))

    def _metrics_json(self):
        metrics = []

        for node in self._nodes:
            metrics.append(dict(name=node.name,
                                input=node.metrics['input'],
                                output=node.metrics['output'],
                                mean_execution_time=f"{mean(node.metrics['execution_time']) * 1000:.2f}",
                                pstdev_execution_time=f"{pstdev(node.metrics['execution_time']) * 1000:.2f}"))

        return dict(current_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'), metrics=metrics)

    def _build_pipeline(self):
        queues = {}

        for node in self._nodes:
            for queue in node.input + node.output:
                if queue in queues:
                    continue

                queues[queue] = Queue(name=queue, node_type=node.node_type, maxsize=self._queue_maxsize)

        for node in self._nodes:
            for queue in node.input:
                node.input_queues.append(queues[queue])

            for queue in node.output:
                node.output_queues.append(queues[queue])

            node.start()

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        self.shared_memory['time'] = time.time()

        if all(node.is_completed for node in self._nodes):
            logging.info(f'{self.name}: AllNodesCompleted')

            self._status = NodeStatus.COMPLETED

        elif any(node.is_terminated for node in self._nodes):
            logging.info(f'{self.name}: AtLeastOneNodeTerminated')

            for node in self._nodes:
                node.drain()
                logging.info(f'{node.name}: NodeDrained')

            self._status = NodeStatus.COMPLETED

        return None



from .core import Batch, BatchStatus, Node, NodeType, Pipeline
from .nodes import NodeFastAPI, NodeONNXRuntime

__all__ = [
    'Batch',
    'BatchStatus',
    'Node',
    'NodeType',
    'Pipeline',
    'NodeFastAPI',
    'NodeONNXRuntime'
]



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import onnxruntime

from ..core.batch import Batch
from ..core.node import Node


class NodeONNXRuntime(Node, ABC):
    def __init__(
            self,
            model_path: str,
            providers: List[Union[str, Tuple[str, Dict]]],
            dtype: np.dtype = np.float32,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._session = onnxruntime.InferenceSession(path_or_bytes=model_path, providers=providers)
        self._dtype = dtype
        self._input_name = self._session.get_inputs()[0].name
        self._batch_size, self._channels, self._height, self._width = self._session.get_inputs()[0].shape

        if not isinstance(self._batch_size, int) or self._batch_size < 1:
            self._batch_size = 32

        self._batch = np.zeros((self._batch_size, self._channels, self._height, self._width), dtype=self._dtype)

    @abstractmethod
    def preprocessing(self, item: Any) -> Tuple[np.ndarray, Any]:
        pass

    @abstractmethod
    def postprocessing(self, output: List[Union[np.ndarray, np.generic]], metadata: Any) -> Any:
        pass

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        if not data:
            return data

        batch = Batch(metadata=data.metadata)
        images, metadatas = zip(*(self.preprocessing(item) for item in data))

        for i in range(0, len(images), self._batch_size):
            for batch_idx, image in enumerate(images[i: i + self._batch_size]):
                self._batch[batch_idx] = image

            outputs = self._session.run(None, {self._input_name: self._batch})

            for *output, metadata in zip(*outputs, metadatas[i: i + self._batch_size]):
                batch.append(self.postprocessing(output, metadata))

        return batch



import logging
import time
from abc import ABC, abstractmethod
from collections import deque
from multiprocessing import Process
from multiprocessing.managers import DictProxy
from threading import Thread
from typing import Dict, List, Optional, Union

from .batch import Batch
from .enums import BatchStatus, NodeStatus, NodeType
from .memory import Memory
from .queue import Queue
from .shared_memory import SharedMemory


class Node(ABC):
    def __init__(
            self,
            node_type: NodeType = NodeType.THREAD,
            input: Optional[Union[str, List[str]]] = None,
            output: Optional[Union[str, List[str]]] = None,
            timeout: float = 0.0
    ):
        self._name = self.__class__.__name__
        self._node_type = node_type
        self._status = NodeStatus.PENDING

        self._input = [] if input is None else [input] if isinstance(input, str) else input
        self._output = [] if output is None else [output] if isinstance(output, str) else output
        self._input_queues = []
        self._output_queues = []

        self._memory = Memory()
        self._shared_memory = SharedMemory.get_shared_memory()
        self._metrics = dict(execution_time=deque([0.0], maxlen=100), input=0, output=0)
        self._timeout = timeout

    @property
    def name(self) -> str:
        return self._name

    @property
    def node_type(self) -> NodeType:
        return self._node_type

    @property
    def is_pending(self) -> bool:
        return self._status == NodeStatus.PENDING

    @property
    def is_active(self) -> bool:
        return self._status == NodeStatus.ACTIVE

    @property
    def is_completed(self) -> bool:
        return self._status == NodeStatus.COMPLETED

    @property
    def is_terminated(self) -> bool:
        return self._status == NodeStatus.TERMINATED

    @property
    def input(self) -> List[str]:
        return self._input

    @property
    def output(self) -> List[str]:
        return self._output

    @property
    def input_queues(self) -> List[Queue]:
        return self._input_queues

    @property
    def output_queues(self) -> List[Queue]:
        return self._output_queues

    @property
    def memory(self) -> Memory:
        return self._memory

    @property
    def shared_memory(self) -> DictProxy:
        return self._shared_memory

    @property
    def metrics(self) -> Dict:
        return self._metrics

    def start(self):
        if self._status != NodeStatus.PENDING:
            raise RuntimeError

        self._status = NodeStatus.ACTIVE

        if self._node_type == NodeType.THREAD:
            Thread(target=self._run, name=self._name).start()
        elif self._node_type == NodeType.PROCESS:
            Process(target=self._run, name=self._name).start()

    def drain(self):
        self._status = NodeStatus.TERMINATED

        for queue in self._input_queues + self._output_queues:
            while not queue.empty():
                queue.get()

    def post_init(self):
        pass

    def release(self):
        pass

    def _run(self):
        self.post_init()

        while self._status == NodeStatus.ACTIVE:
            if not self._input_queues:
                input = None
            elif len(self._input_queues) == 1:
                input = self._input_queues[0].get()
            elif len(self._input_queues) > 1:
                batches = [queue.get() for queue in self._input_queues]

                if any(len(batch) != len(batches[0]) for batch in batches):
                    logging.error(f'{self.name}: BatchLengthMismatchError')
                    self._status = NodeStatus.TERMINATED
                    break

                if all(batch.is_ok for batch in batches):
                    status = BatchStatus.OK
                elif all(batch.is_last for batch in batches):
                    status = BatchStatus.LAST
                else:
                    logging.error(f'{self.name}: BatchStatusMismatchError')
                    self._status = NodeStatus.TERMINATED
                    break

                input = Batch(status=status)

                for idx in range(len(batches[0])):
                    input.append({
                        queue.name: batch[idx]
                        for queue, batch in zip(self._input_queues, batches)
                    })

                for batch in batches:
                    input.metadata.update(batch.metadata)

            try:
                start_time = time.perf_counter()
                output = self.processing(input)
                end_time = time.perf_counter()
                self._metrics['execution_time'].append(end_time - start_time)
                self._metrics['input'] += len(input) if isinstance(input, Batch) else 0
                self._metrics['output'] += len(output) if isinstance(output, Batch) else 0
            except Exception as e:
                logging.error(f'{self.name}: {e.__class__} {e}', exc_info=True)
                self._status = NodeStatus.TERMINATED
                break

            if (
                (output is None and self._output_queues) or
                (isinstance(output, Batch) and not self._output_queues)
            ):
                logging.error(f'{self.name}: NodeOutputMismatchError')
                self._status = NodeStatus.TERMINATED
                break

            if (
                isinstance(input, Batch) and isinstance(output, Batch) and
                (
                    (input.is_ok and output.is_last) or
                    (input.is_last and output.is_ok)
                )
            ):
                logging.error(f'{self.name}: BatchStatusMismatchError')
                self._status = NodeStatus.TERMINATED
                break

            if isinstance(output, Batch):
                for queue in self._output_queues:
                    queue.put(output)

            if (
                (isinstance(input, Batch) and input.is_last) or
                (isinstance(output, Batch) and output.is_last)
            ):
                self._status = NodeStatus.COMPLETED
                break

            time.sleep(self._timeout)

        self.release()

    @abstractmethod
    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        pass
